/*
 * Copyright (c) 2015 N. Billard.
 */
package com.massa.mrules.demos.eshop;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.compile.MCompilationContext;
import com.massa.mrules.context.execute.MExecutionContext;
import com.massa.mrules.demos.eshop.model.Article;
import com.massa.mrules.demos.eshop.model.ArticleType;
import com.massa.mrules.demos.eshop.model.Basket;
import com.massa.mrules.demos.eshop.model.Command;
import com.massa.mrules.demos.eshop.model.CommandHistory;
import com.massa.mrules.demos.eshop.model.Customer;
import com.massa.mrules.demos.eshop.model.CustomerLevel;
import com.massa.mrules.exception.MActionExecutionException;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MExecutionException;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.RichXMLFactory;
import com.massa.mrules.set.IMRuleExecutionSet;

@RunWith(Parameterized.class)
public class EshopTest {
	private static final Log LOG = LogFactory.getLog(EshopTest.class.getName());

	final CompilationLevel level;

	public EshopTest(final CompilationLevel level) {
		this.level = level;
	}

	@Parameters
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
			{CompilationLevel.STANDARD},
			{CompilationLevel.LIGHT}
		});
	}

	private static final String XML = "com/massa/mrules/demos/eshop/rules.xml";

	@Test
	public void eshopDemo() throws MConfigurationException, MExecutionException, IOException {
		final List<Basket> manualBaskets = this.buildTestBasket();
		final List<Basket> engineBaskets = this.buildTestBasket();

		final RichXMLFactory factory = new RichXMLFactory(new MCompilationContext(this.level, Basket.class));

		final InputStream xml = ClassLoader.getSystemResourceAsStream(EshopTest.XML);
		final IMRuleExecutionSet eng = factory.read(xml);

		@SuppressWarnings("unchecked")
		final
		MExecutionContext<Basket> ctxt = (MExecutionContext<Basket>) eng.getContextFactory().getExecutionContext();

		for (final Basket basket: engineBaskets) {
			try {
				ctxt.setInput(basket);
				ctxt.execute();
				if (basket.isShouldBeInError()) {
					Assert.fail("Error not detected.");
				}
			} catch (final MActionExecutionException e) {
				System.out.println(e.getMessage());
				if (basket.isError() && ! basket.isShouldBeInError()) {
					Assert.fail("Error abusively detected.");
				}
				if (! basket.isError() && basket.isShouldBeInError()) {
					Assert.fail("Error flag not set.");
				}
				if (! "RAISEERROR".equals(e.getInfoBlock().getMessageInfo().getMessageCode().getErrorCode())) {
					e.printStackTrace();
					Assert.fail("Error flag not set or wrong exception thrown.");
				}
			}
		}

		this.manualCompute(manualBaskets);

		this.compare(manualBaskets, engineBaskets);

		//Time check
		EshopTest.LOG.info("Beginning performance test.");
		ctxt.setLoggerName(".perf"); //to avoid debug and info.
		ctxt.setOut(null); //To avoid thousands of printings.
		engineBaskets.clear();
		for (int i = 0 ; i < 10000 ; i++) {
			engineBaskets.addAll(this.buildTestBasket());
		}

		final long t1 = System.currentTimeMillis();
		for (final Basket basket: engineBaskets) {
			try {
				ctxt.setInput(basket);
				ctxt.execute();
			} catch (final MActionExecutionException e) {
			}
		}
		EshopTest.LOG.info("Time for 10000 tests : " + (System.currentTimeMillis() - t1));

		//Read / Write tests.
		EshopTest.LOG.info("Human readable version:\n" + eng.toString());

		final StringWriter w = new StringWriter();
		factory.write(w, eng);
		EshopTest.LOG.info(w.toString());

		final ByteArrayOutputStream o = new ByteArrayOutputStream();
		final SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, eng);

		Assert.assertEquals(eng, eng.clone());
		Assert.assertEquals(eng, factory.read(new StringReader(w.toString())));
		Assert.assertEquals(eng, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}

	private void manualCompute(final List<Basket> baskets) {
		int idxdiscounts = 0;
		basketsLabel:
			for (final Basket basket: baskets) {
				// Computing customer level.
				double historicalAmounts = 0.0D;
				int historicalCommands = 0;

				if (basket.getCustomer().getHistory() != null) {
					for (final CommandHistory ch: basket.getCustomer().getHistory()) {
						historicalAmounts += ch.getFinalAmount();
						if (ch.getFinalAmount() >= 500) {
							historicalCommands ++;
						}
					}
				}

				if (historicalAmounts >= 1500) {
					basket.getCustomer().setLevel(CustomerLevel.GOLD);
				} else if (historicalAmounts >= 1000 || historicalCommands > 0) {
					basket.getCustomer().setLevel(CustomerLevel.SILVER);
				} else {
					basket.getCustomer().setLevel(CustomerLevel.STANDARD);
				}

				double sumNetPrices = 0.0D;
				double sumVat = 0.0D;
				//Computing Articles TVA and global prices.
				for (final Article article: basket.getCommand().getArticles()) {
					article.setVat(article.getType() == ArticleType.FOOD ? 5.5D : 20.0D);
					sumNetPrices += article.getPrice();
					sumVat += article.getPrice() * article.getVat() / 100.0D;
				}
				basket.getCommand().setTotalNetAmount(Math.round(100 * sumNetPrices) / 100.0D);
				basket.getCommand().setTotalVat(Math.round(100 * sumVat) / 100.0D);
				basket.getCommand().setTotalAmount(Math.round(100 * (sumNetPrices + sumVat)) / 100.0D);

				int idx = 0;
				double sumDiscounts = 0.0D;
				if (basket.getCommand().getUsedDiscounts() != null) {
					for (final String discountCode: basket.getCommand().getUsedDiscounts()) {
						++idx;
						if (basket.getCustomer().getDiscounts() == null || basket.getCustomer().getDiscounts().get(discountCode) == null) {
							System.out.println("Manual : Discount code number " + idx + " is not valid.");
							continue basketsLabel;
						}
						sumDiscounts += basket.getCommand().getTotalAmount() * basket.getCustomer().getDiscounts().get(discountCode) / 100.0D;
					}
				}

				//Special discount if at least 2 of [deliveryMode:  ECONOMIC | deliveryCity: PARIS | level: GOLD] : 15 CURRENCY
				int specialNum = 0;
				if ("ECONOMIC".equals(basket.getCommand().getDeliveryMode())) {
					specialNum++;
				}
				if ("PARIS".equals(basket.getCommand().getDeliveryCity())) {
					specialNum++;
				}
				if (CustomerLevel.GOLD == basket.getCustomer().getLevel()) {
					specialNum++;
				}
				if (specialNum >= 2) {
					basket.getCommand().setSpecialDiscount(true);
					sumDiscounts += 15;
				}

				basket.getCommand().setTotalDiscount(Math.round(100 * sumDiscounts) / 100.0D);
				basket.getCommand().setFinalAmount(basket.getCommand().getTotalAmount() - basket.getCommand().getTotalDiscount());

				//Command > 500 --> 10% future discount
				if (basket.getCommand().getFinalAmount() >= 500) {
					if (basket.getCustomer().getDiscounts() == null) {
						basket.getCustomer().setDiscounts(new HashMap<String, Short>());
					}
					basket.getCustomer().getDiscounts().put("GENERATED_DISCOUNT" + (++idxdiscounts), Short.valueOf((short) 10));
				}

				//5th command --> 5% future discount
				if (basket.getCustomer().getHistory() != null && basket.getCustomer().getHistory().size() == 4) {
					basket.getCustomer().getDiscounts().put("GENERATED_DISCOUNT" + (++idxdiscounts), Short.valueOf((short) 5));
				}
			}
	}

	private void compare(final List<Basket> manualBaskets, final List<Basket> engineBaskets) {
		for (int i = 0 ; i < manualBaskets.size() ; i++) {
			final Basket manualBasket = manualBaskets.get(i);
			final Basket engineBasket = engineBaskets.get(i);

			Assert.assertEquals("index : " + i, manualBasket.getCustomer().getLevel(), engineBasket.getCustomer().getLevel());
			Assert.assertEquals("index : " + i, manualBasket.getCommand().getTotalNetAmount(), engineBasket.getCommand().getTotalNetAmount(), 0.00001D);
			Assert.assertEquals("index : " + i, manualBasket.getCommand().getTotalVat(), engineBasket.getCommand().getTotalVat(), 0.00001D);
			Assert.assertEquals("index : " + i, manualBasket.getCommand().getTotalAmount(), engineBasket.getCommand().getTotalAmount(), 0.00001D);
			Assert.assertEquals("index : " + i, manualBasket.getCommand().getTotalDiscount(), engineBasket.getCommand().getTotalDiscount(), 0.00001D);
			Assert.assertEquals("index : " + i, manualBasket.getCommand().getFinalAmount(), engineBasket.getCommand().getFinalAmount(), 0.00001D);
			Assert.assertEquals("index : " + i, manualBasket.getCommand().isSpecialDiscount(), engineBasket.getCommand().isSpecialDiscount());
		}
	}

	private List<Basket> buildTestBasket() {
		final List<Basket> baskets = new ArrayList<Basket>();

		Basket basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[3]);
		basket.getCommand().getArticles()[0] = new Article("Television", ArticleType.ELECTRONIC, 349.90);
		basket.getCommand().getArticles()[1] = new Article("HI-FI system", ArticleType.ELECTRONIC, 389.49);
		basket.getCommand().getArticles()[2] = new Article("Television support", ArticleType.DECORATION, 139.99);
		basket.getCommand().setUsedDiscounts(new String[] {"DISCOUNT_1"});
		basket.getCustomer().getHistory().add(new CommandHistory(5, 158.26));
		basket.getCustomer().getHistory().add(new CommandHistory(3, 227.84));
		basket.getCustomer().getHistory().add(new CommandHistory(5, 158.18));
		basket.getCustomer().getDiscounts().put("DISCOUNT_1", Short.valueOf((short) 5));
		basket.getCustomer().getDiscounts().put("DISCOUNT_2", Short.valueOf((short) 8));
		basket.getCommand().setDeliveryCity("PARIS");
		basket.getCommand().setDeliveryMode("FAST");

		basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[1]);
		basket.getCommand().getArticles()[0] = new Article("Refrigerator", ArticleType.ELECTRONIC, 595);
		basket.getCustomer().setDiscounts(null);
		basket.getCustomer().setHistory(null);
		basket.getCommand().setDeliveryCity("PARIS");
		basket.getCommand().setDeliveryMode("ECONOMIC");

		basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[3]);
		basket.getCommand().getArticles()[0] = new Article("Computer", ArticleType.ELECTRONIC, 540.99);
		basket.getCommand().getArticles()[1] = new Article("Screen", ArticleType.ELECTRONIC, 217.29);
		basket.getCommand().getArticles()[2] = new Article("Orange juice", ArticleType.FOOD, 9.99);
		basket.getCommand().setUsedDiscounts(new String[] {"DISCOUNT_1", "DISCOUNT_3"});
		basket.getCustomer().getHistory().add(new CommandHistory(12, 853.27));
		basket.getCustomer().getHistory().add(new CommandHistory(2, 380.12));
		basket.getCustomer().getHistory().add(new CommandHistory(1, 25.89));
		basket.getCustomer().getHistory().add(new CommandHistory(1, 85.40));
		basket.getCustomer().getDiscounts().put("DISCOUNT_1", Short.valueOf((short) 10));
		basket.getCustomer().getDiscounts().put("DISCOUNT_2", Short.valueOf((short) 3));
		basket.getCustomer().getDiscounts().put("DISCOUNT_3", Short.valueOf((short) 7));
		basket.getCommand().setDeliveryCity("BORDEAUX");
		basket.getCommand().setDeliveryMode("FAST");

		basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[1]);
		basket.getCommand().getArticles()[0] = new Article("Expensive Article", ArticleType.ELECTRONIC, 1000.99);
		basket.getCommand().setUsedDiscounts(new String[] {"DISCOUNT_1"});
		basket.getCustomer().getHistory().add(new CommandHistory(4, 500.32));
		basket.getCustomer().getHistory().add(new CommandHistory(3, 856.82));
		basket.getCustomer().getHistory().add(new CommandHistory(5, 432.49));
		basket.getCustomer().getDiscounts().put("DISCOUNT_1", Short.valueOf((short) 3));
		basket.getCommand().setDeliveryCity("BORDEAUX");
		basket.getCommand().setDeliveryMode("ECONOMIC");

		//With error on discount validity.
		basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[1]);
		basket.getCommand().getArticles()[0] = new Article("Radio", ArticleType.ELECTRONIC, 149.95);
		basket.getCommand().setUsedDiscounts(new String[] {"DISCOUNT_1", "DISCOUNT_INVALID"});
		basket.getCustomer().setHistory(null);
		basket.getCustomer().getDiscounts().put("DISCOUNT_1", Short.valueOf((short) 10));
		basket.setShouldBeInError(true);
		basket.getCommand().setDeliveryCity("PARIS");
		basket.getCommand().setDeliveryMode("ECONOMIC");

		//With error on discount validity and null discounts.
		basket = this.buildEmptyTestBasket();
		baskets.add(basket);
		basket.getCommand().setArticles(new Article[1]);
		basket.getCommand().getArticles()[0] = new Article("Radio", ArticleType.ELECTRONIC, 149.95);
		basket.getCommand().setUsedDiscounts(new String[] {"DISCOUNT_INVALID"});
		basket.getCustomer().setHistory(null);
		basket.getCustomer().setDiscounts(null);
		basket.setShouldBeInError(true);
		basket.getCommand().setDeliveryCity("PARIS");
		basket.getCommand().setDeliveryMode("ECONOMIC");

		return baskets;
	}

	private Basket buildEmptyTestBasket() {
		final Basket basket = new Basket();

		basket.setCommand(new Command());
		basket.setCustomer(new Customer());

		basket.getCustomer().setHistory(new ArrayList<CommandHistory>());
		basket.getCustomer().setDiscounts(new HashMap<String, Short>());

		return basket;
	}
}
